# Any changes to the anomaly library should be reinstalled with
#  pip install --upgrade .

# For running unit tests, use
# /usr/bin/python -m unittest test


import unittest

from anomaly import Anomaly


import pandas as pd
df = pd.DataFrame({'ano_mes' :  ['2020_jan', '2020_fev','2020_mar', '2020_abr','2020_mai','2020_jun','2020_jul',
                         '2020_ago','2020_set','2020_out','2020_nov','2020_dez','2021_jan'],
           'val_agg' :  [10, 12, 8, 15, 17, 15, 17, 20, 23, 18, 20, 20, 25],
           'tipo_dado': ['historical','historical','historical','historical','historical',
                         'historical','historical','historical','historical','historical',
                         'historical', 'historical','target']})


class TestAnomalyClass(unittest.TestCase):
    def setUp(self):
        self.anomaly = Anomaly(_index = df['ano_mes'], _values = df['val_agg'], _type = df['tipo_dado'])

    def test_meancalculation(self):
        self.anomaly.calculate_mean()
        
        mean_pandas = df[df['tipo_dado'] == 'historical']['val_agg'].mean()
        
        self.assertEqual(self.anomaly.mean, mean_pandas, 'incorrect mean')
        
    def test_stdevcalculation(self):
        self.anomaly.calculate_stdev()
        
        stdev_pandas =  df[df['tipo_dado'] == 'historical']['val_agg'].std()
        
        self.assertEqual(self.anomaly.stdev, stdev_pandas, 'incorrect standard deviation')
        
    
    def test_confidenceinterval(self):
        self.anomaly.calculate_mean()
        self.anomaly.calculate_stdev()
        self.anomaly.calculate_confidence_interval()
        
        mean_pandas = df[df['tipo_dado'] == 'historical']['val_agg'].mean()
        stdev_pandas =  df[df['tipo_dado'] == 'historical']['val_agg'].std()
        
        
        lims = [1, 2, 3]
        
        for i in lims:
            
            lower_limit_pandas = mean_pandas - i*stdev_pandas
            higher_limit_pandas = mean_pandas + i*stdev_pandas 

            if lower_limit_pandas < 0:
                lower_limit_pandas = 0

            self.assertEqual(self.anomaly.CI[i-1][0], lower_limit_pandas, 'incorrect confidence interval {}sd lower limit ')
            self.assertEqual(self.anomaly.CI[i-1][1], higher_limit_pandas, 'incorrect confidence interval {}sd higher limit ')

    def test_target(self):        
        target_pandas =  df[df['tipo_dado'] == 'target']['val_agg'].values[0]
        
        self.anomaly.obtain_target()
        self.assertEqual(self.anomaly.target, target_pandas, 'target did not match')
        
    def test_anomalydetection(self):
        self.anomaly.calculate_mean()
        self.anomaly.calculate_stdev()
        self.anomaly.calculate_confidence_interval()  
        self.anomaly.obtain_target()
        self.anomaly.identify_anomaly()
        
        
        lims = [1, 2, 3]
        
        for i in lims:
            print('>>> CI_{}sd {}'.format(i, self.anomaly.CI[i-1]))
            
            results = ['Anomaly', 'Not_anomaly', 'Not_anomaly']
            self.assertEqual(self.anomaly.result[i-1], results[i-1], 'anomaly not detected')


if __name__ == '__main__':
    unittest.main()


#%%writefile anomaly.py

import statistics

class Anomaly():
    '''
    Anomaly detector under statistical method
    
    Attributes:
       _index (obj): list of year_month
       _values (float): agregated values to detect
       _type (obj): list identifying values as: 'historical' or 'target'
       
       
    '''
    def __init__(self, _index, _values, _type, _limits = [1, 2, 3]):
        
        self.values = _values
        self.index = _index
        self.type = _type
        
        self.mean = 0
        self.stdev = 0
        
        self.target = 0
        
        self.limits = _limits

        self.CI = []        

        self.result = []
        
    def calculate_mean(self):
        
        self.mean = statistics.mean(self.values[self.type == 'historical'])
        
        return self.mean
    
    
    def calculate_stdev(self):
        
        self.stdev = statistics.stdev(self.values[self.type == 'historical'])
        
        return self.stdev
    
    
    def calculate_confidence_interval(self):
        ''' 
        Calculate confidence interval iterating over the limits list.
        
        eg. limit = 1 :> superior limit = mean + 1 times standard deviation
        '''
        for limit in self.limits:
            
            lim_inf = self.mean - self.stdev*limit
            lim_sup = self.mean + self.stdev*limit
                
            if lim_inf < 0: # Zeroing negative lower confidence interval
                lim_inf = 0
        
            self.CI.append([lim_inf, lim_sup])

        return self.CI
    
    def obtain_target(self):
        
        self.target = self.values[self.type == 'target'].values[0]
        self.target_date = self.index[self.type == 'target'].values[0]
    
    
    def identify_anomaly(self):
        'Anomaly based on superior limit only in v0.01'
        self.result = []
                
        for i in range(0, len(self.limits)): #Iterates over limits number
        
            if self.target > self.CI[i][1]:            
                self.result.append('Anomaly')
            else:            
                self.result.append('Not_anomaly')

        return self.result
    
    
    def print_data(self):
        import pandas as pd
        print('-------------------')
        data = pd.DataFrame({'vl_agg' : self.values,
                             'index' : self.index,
                             'type': self.type})
        print(data)
        
        return data



from setuptools import setup

setup(name='simple_anomaly',
      version='0.1',
      description='Anomaly detection',
      packages=['simple_anomaly'],
      author_email='rodrigosannn@gmail.com',
      zip_safe=False)
